#!/usr/bin/env python3

"""NiFi worker to pass Scio produced data to the ACT platform"""

import argparse
import json
import os
import sys
import traceback
from logging import error
from typing import Dict, List, Text, cast

import act.api
from act.api.helpers import handle_fact, handle_uri
from act.workers.libs import worker

EXTRACT_GEONAMES = ["countries", "regions", "regions-derived",
                    "sub-regions", "sub-regions-derived"]


EXTRACT_INDICATORS = ["md5", "sha1", "sha256",
                      "fqdn", "ipv4", "ipv6", "email",
                      "msid", "cve", "uri", "ipv4net"]

SCIO_GEONAMES_ACT_MAP = {
    "countries": "country",
    "regions": "region",
    "regions-derived": "region",
    "sub-regions": "subRegion",
    "sub-regions-derived": "subRegion"
}

SCIO_INDICATOR_ACT_MAP = {
    "md5": "hash",
    "sha1": "hash",
    "sha256": "hash",
    "ipv4net": "ipv4Network",
    "cve": "vulnerability",
    "msid": "vulnerability",
}


def parseargs() -> argparse.ArgumentParser:
    """ Parse arguments """
    parser = worker.parseargs('Get SCIO reports and IOCs from stdin')
    return parser


def get_scio_report() -> Dict:
    """Read scio report from stdin"""

    return cast(Dict, json.load(sys.stdin))


def report_mentions_fact(actapi: act.api.Act, object_type: Text, object_values: List[Text], report_id: Text, output_format: Text) -> None:
    """Add mentions fact to report"""
    for value in list(set(object_values)):
        try:
            handle_fact(
                actapi.fact("mentions", object_type)
                .source("report", report_id)
                .destination(object_type, value),
                output_format
            )
        except act.api.base.ResponseError as e:
            error("Unable to create linked fact: %s" % e)


def add_to_act(actapi: act.api.Act, doc: Dict, output_format: Text = "json") -> None:
    """Add a report to the ACT platform"""

    report_id: Text = doc["hexdigest"]
    title: Text = doc.get("title", "No title")
    indicators: Dict = doc.get("indicators", {})

    try:
        # Report title
        handle_fact(
            actapi.fact("name", title)
            .source("report", report_id),
            output_format
        )
    except act.api.base.ResponseError as e:
        error("Unable to create fact: %s" % e)

    # Loop over all items under indicators in report
    for scio_indicator_type in EXTRACT_INDICATORS:
        # Get object type from ACT (default to object type in SCIO)
        act_indicator_type = SCIO_INDICATOR_ACT_MAP.get(scio_indicator_type,
                                                        scio_indicator_type)
        report_mentions_fact(
            actapi,
            act_indicator_type,
            indicators.get(scio_indicator_type, []),
            report_id,
            output_format)

    # For SHA256, create content object
    for sha256 in list(set(indicators.get("sha256", []))):
        handle_fact(
            actapi.fact("represents")
            .source("hash", sha256)
            .destination("content", sha256),
            output_format
        )

    # Add all URI components
    for uri in list(set(indicators.get("uri", []))):
        try:
            handle_uri(actapi, uri, output_format=output_format)
        except act.api.schema.MissingField:
            error("Unable to create facts from uri: {}".format(uri))

    # Locations (countries, regions, sub regions)
    for location_type in EXTRACT_GEONAMES:
        locations = doc.get("geonames", {}).get(location_type, [])

        report_mentions_fact(
            actapi,
            SCIO_GEONAMES_ACT_MAP[location_type],
            locations,
            report_id,
            output_format)

    # Threat actor
    report_mentions_fact(
        actapi,
        "threatActor",
        doc.get("threat-actor", {}).get("names", []),
        report_id,
        output_format)

    # Tools
    report_mentions_fact(
        actapi,
        "tool",
        [tool.lower() for tool in doc.get("tools", {}).get("names", [])],
        report_id,
        output_format)

    # Sector
    report_mentions_fact(
        actapi,
        "sector",
        doc.get("sectors", []),
        report_id,
        output_format)


def main() -> None:
    """main function"""

    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = worker.handle_args(parseargs())

    actapi = worker.init_act(args)

    # Add IOCs from reports to the ACT platform
    add_to_act(
        actapi,
        get_scio_report(),
        args.output_format,
    )


def main_log_error() -> None:
    """Execute main() and log errors to error"""
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == '__main__':
    main_log_error()



#!/usr/bin/env python3

"""General ACT backend uploader. Reads facts as JSON
from the stdin, uploading accordingly"""

import json
import os
import sys
import traceback
from logging import error

import act.api
from act.workers.libs import worker


def main(actapi: act.api.Act) -> None:
    """Process stdin, parse each separat line as a JSON structure and
    register a fact based on the structure. The form of input should
    be the on the form accepted by the ACT Rest API fact API."""

    for line in sys.stdin:
        data = json.loads(line)

        fact = actapi.fact(**data)
        try:
            fact.add()
        except act.api.base.ResponseError as err:
            error("ResponseError while storing objects: %s" % err)


def main_log_error() -> None:
    "Call main() and log all exceptions as errors"
    try:
        # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
        # (or replace .config with $XDG_CONFIG_DIR if set)
        args = worker.handle_args(worker.parseargs("Generic uploader"))
        actapi = worker.init_act(args)

        main(actapi)
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == '__main__':
    main_log_error()



#!/usr/bin/env python3
import json
import os
import argparse
from act.workers.libs.ta_list_methods import get_all_ta_from_act, get_all_alias_facts_from_act, \
    add_ta_to_map, add_ta_alias_to_map, create_config

from act.workers.libs import worker

def parseargs() -> argparse.ArgumentParser:
    """ Parse arguments """
    parser = worker.parseargs('Updates list of threat actor aliases from an ACT platform')
    parser.add_argument('--aliasfile', required=True, help=".cfg-file with \
        existing threat actor aliases")
    parser.add_argument('--newaliasfile', required=True, help="name of \
        .cfg-file with updated list of threat actor aliases")
    parser.add_argument('--output-json', action='store_true', help="Enable \
        this flag if json testfiles should be created.")

    return parser

def main():
    """main function"""

    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = worker.handle_args(parseargs())

    # gets all ta names from objects(as a set of strings) and facts(as a
    # set with tuples of two strings) in ACT.

    threatactors = get_all_ta_from_act(args.baseurl, args.userid)

    # gets all threat actor aliases from act platform

    ta_aliases = get_all_alias_facts_from_act(args.baseurl, args.userid)

    # save ta and ta_aliases to json test file
    if args.output_json:
        with open('test/objects.json', 'w') as outfile:
            outfile.write(json.dumps(list(threatactors)))

        with open('test/facts.json', 'w') as outfile:
            outfile.write(json.dumps(list(ta_aliases)))

    # adds all ta names from threatActor objects from ACT into ta_map
    ta_map = add_ta_to_map(threatactors)

    # adds all ta names from alias-facts in ACT to the ta_map
    ta_map_with_aliases = add_ta_alias_to_map(ta_aliases, ta_map)

    # creates a new .cfg-file including all the content from ta_map
    # and the existing cfg.-file from arguments.
    create_config(ta_map_with_aliases, args.aliasfile, args.newaliasfile)


if __name__ == "__main__":

    main()



import enum
import uuid
import json
import time
import re
import ipaddress

from typing import Optional, Text, Any, Tuple, Dict, Callable

class ThreatLevelID(enum.Enum):
    """2.2.1.5.  threat_level_id

    threat_level_id represents the threat level.

    4:   Undefined
    3:   Low
    2:   Medium
    1:   High

    If a higher granularity is required, a MISP taxonomy applied as a Tag
    SHOULD be preferred.

    threat_level_id SHALL be present."""

    HIGH = 1
    MEDIUM = 2
    LOW = 3
    UNDEFINED = 4


class Analysis(enum.Enum):
    """2.2.1.6.  analysis

    analysis represents the analysis level.

    0:   Initial
    1:   Ongoing
    2:   Complete

    If a higher granularity is required, a MISP taxonomy applied as a Tag
    SHOULD be preferred.

    analysis SHALL be present."""

    INITIAL = 0
    ONGOING = 1
    COMPLETE = 2


class Distribution(enum.Enum):
    """2.2.1.13.  distribution

    distribution represents the basic distribution rules of the event.  The
    system must adhere to the distribution setting for access control and for
    dissemination of the event.

    distribution MUST be present and be one of the following options:

    0 Your Organisation Only
    1 This Community Only
    2 Connected Communities
    3 All Communities
    4 Sharing Group"""

    ORGANIZATION_ONLY = 0
    COMMUNITY_ONLY = 1
    CONNECTED_COMMUNITIES = 2
    ALL_COMMUNITIES = 3
    SHARING_GROUP = 4
    INHERIT_EVENT = 5


class Event(object):

    # --- MUST ----
    _uuid = None
    published = None
    info = None
    threat_level_id = ThreatLevelID.UNDEFINED
    analysis = Analysis.INITIAL
    date = None  # ISO-8601 (date only: YYYY-MM-DD) reference date of the event
    timestamp = None  # Timestamp of event creation or event/attribute last update
    publish_timestamp = None  # reference time when the event was published on the instance.
    org_id = None  # Human readable org. generating the Event
    orgc_id = None  # Human readble org. *creating* the Event
    attribute_count = 0
    distribution = Distribution.ORGANIZATION_ONLY
    sharing_group_id = None
    # --- END_MUST ----

    # --- SHOULD ---
    extends_uuid = None
    # --- END_SHOULD ---

    tlp = None
    misp_galaxy: Dict[Text, Text] = {}

    def __init__(self, loads: Optional[Text]=None) -> None:
        self._uuid: uuid.UUID = uuid.uuid4()
        self.timestamp: int = int(time.time())
        self.publish_timestamp: int = int(time.time())

        if loads:
            data = json.loads(loads)
            event = data["Event"]
            self._uuid = uuid.UUID(event["uuid"])
            self.published = event["published"]
            self.info = event["info"]
            self.threat_level_id = ThreatLevelID(int(event["threat_level_id"]))
            self.analysis = Analysis(int(event["analysis"]))
            self.date = event["date"]
            self.timestamp = event["timestamp"]
            self.publish_timestamp = event["publish_timestamp"]
            self.org_id = event.get("org_id", "N/A")
            self.orgc_id = event.get("orgc_id", "N/A")
            self.attribute_count = event.get("attribute_count", 0)
            self.distribution = Distribution(int(event.get("distribution", 0)))
            self.sharing_group_id = event.get("sharing_group_id", None)
            self.extends_uuid = event.get("extends_uuid", None)

            misp_re = re.compile(r'misp-galaxy:(.*?)="(.*?)"')
            for tag in event.get("Tag", []):
                name = tag["name"]
                if name.startswith("tlp"):
                    self.tlp = name.split(":")[1]
                if name.startswith("misp-galaxy"):
                    for match in misp_re.findall(name):
                        self.misp_galaxy[match[0]] = match[1]

            self.attributes = [Attribute(e) for e in event.get("Attribute", [])]
            objects = event.get("Object", [])
            for obj in objects:
                obj_attributes = obj.get("Attribute", [])
                self.attributes += [Attribute(e) for e in obj_attributes]

    def __str__(self) -> Text:
        return "({0}) {1} - {2} ".format(self.timestamp, self._uuid, self.info)

    def write_to(self, stream: Any) -> None:
        stream.write(json.dumps(
            {
                "uuid": str(self._uuid),
                "published": self.published,
                "info": self.info,
                "threat_level_id": self.threat_level_id.value,
                "analysis": self.analysis.value,
                "date": self.date,
                "timestamp": self.timestamp,
                "publish_timestamp": self.publish_timestamp,
                "org_id": self.org_id,
                "orgc_id": self.orgc_id,
                "attribute_count": self.attribute_count,
                "distribution": self.distribution.value,
                "sharing_group_id": self.sharing_group_id,
                "extends_uuid": self.extends_uuid,
            }))

    @property
    def uuid(self):  # type: ignore
        return self._uuid


class Attribute(object):  # attributeattributes in misp babel

    def __init__(self, attributedict: Dict[Text, Text]):

        try:
            self._uuid = attributedict["uuid"]
            self.id: Text = attributedict["uuid"]
            mapper_fn = map_misp_to_act.get(attributedict["type"], lambda x: (None, None))
            self.act_type: Optional[Text] = None
            self.value: Optional[Text] = None
            self.act_type, self.value = mapper_fn(attributedict["value"])
            if "RelatedAttribute" in attributedict and attributedict["RelatedAttribute"]:
                print("DEBUG: {0}".format(attributedict["RelatedAttribute"]))
        except:
            print(attributedict)
            print(attributedict["value"][:100])
            raise

    def __str__(self) -> Text:
        return "{0} {1}:{2}".format(self.id, self.act_type, self.value)


def hash_f(x: Text) -> Tuple[Text, Text]:
    return "hash", x.lower()


def certificate_f(x: Text) -> Tuple[Text, Text]:
    return "certificate", x.lower()


def threat_actor_f(x: Text) -> Tuple[Text, Text]:
    return "threatActor", x.lower()


def campaign_f(x: Text) -> Tuple[Text, Text]:
    return "campaign", x.lower()


def email_f(x: Text) -> Tuple[Text, Text]:
    return "email", x.lower()


def person_f(x: Text) -> Tuple[Text, Text]:
    return "person", x.lower()


def organization_f(x: Text) -> Tuple[Text, Text]:
    return "organization", x.lower()


def fqdn_f(x: Text) -> Tuple[Text, Text]:
    return "fqdn", x.lower()


def ip_f(x: Text) -> Tuple[Optional[Text], Optional[Text]]:
    try:
        addrv6 = ipaddress.IPv6Address(x)
        return "ipv6", str(addrv6.exploded)
    except ipaddress.AddressValueError:
        try:
            ipaddress.IPv4Address(x)
            return "ipv4", x
        except ipaddress.AddressValueError:
            pass

    return None, None


def uri_f(x: Text) -> Tuple[Text, Text]:
    if not x.startswith("http"):
        x = "http://{0}".format(x)
    return "uri", x


def user_agent_f(x: Text) -> Tuple[Text, Text]:
    return "userAgent", x


def vulnerability_f(x: Text) -> Tuple[Text, Text]:
    return "vulnerability", x.lower()


def mutex_f(x: Text) -> Tuple[Text, Text]:
    return "mutex", x


map_misp_to_act: Dict[Text, Callable[[Text], Tuple[Optional[Text], Optional[Text]]]] = {
    "authentihash": hash_f,
    "campaign-name": campaign_f,
    "hostname": fqdn_f,
    "domain": fqdn_f,
    "impfuzzy": hash_f,
    "imphash": hash_f,
    "ip-dst": ip_f,
    "ip-src": ip_f,
    "link": uri_f,
    "md5": hash_f,
    "mutex": mutex_f,
    "sha1": hash_f,
    "sha224": hash_f,
    "sha256": hash_f,
    "sha384": hash_f,
    "sha512/224": hash_f,
    "sha512/256": hash_f,
    "sha512": hash_f,
    "ssdeep": hash_f,
    "threat-actor": threat_actor_f,
    "url": uri_f,
    "user-agent": user_agent_f,
    "vulnerability": vulnerability_f,
    "whois-registrant-email": email_f,
    "whois-registrant-name": person_f,
    "whois-registrar": organization_f,
    "x509-fingerprint-sha1": certificate_f,
}


class IncompleteEventException(Exception):
    pass



#!/usr/bin/env python3
""" Handle worker config"""

import argparse
import os
import sys
from typing import Text

from pkg_resources import resource_string

from act.workers.libs import config, worker


def parseargs() -> argparse.Namespace:
    """ Parse arguments """

    parser = argparse.ArgumentParser("ACT worker config", epilog="""
    show - Print default config

    user - Copy default config to {0}/{1}

    system - Copy default config to /etc/{1}

""".format(config.get_config_dir(worker.CONFIG_ID), worker.CONFIG_NAME), formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('action', nargs=1, choices=["show", "user", "system"])

    return parser.parse_args()


def default_ini() -> Text:
    "Get content of default ini file"
    return resource_string("act.workers", "etc/{}".format(worker.CONFIG_NAME)).decode('utf-8')


def save_config(filename: Text) -> None:
    """ Save config to specified filename """
    if os.path.isfile(filename):
        sys.stderr.write("Config already exists: {}\n".format(filename))
        sys.exit(1)

    try:
        with open(filename, "w") as f:
            f.write(default_ini())
    except PermissionError as err:
        sys.stderr.write("{}\n".format(err))
        sys.exit(2)

    print("Config copied to {}".format(filename))


def main() -> None:
    "main function"
    args = parseargs()

    if "show" in args.action:
        print(default_ini())

    if "user" in args.action:
        config_dir = config.get_config_dir(worker.CONFIG_ID, create=True)
        save_config(os.path.join(config_dir, worker.CONFIG_NAME))

    if "system" in args.action:
        save_config("/etc/{}".format(worker.CONFIG_NAME))


if __name__ == '__main__':
    main()



""" Common functions towards mnemonic API """

from logging import debug
from typing import Any, Dict, Generator, Optional, Text

import requests

from . import worker


def batch_query(
        method: Text,
        url: Text,
        headers: Optional[Dict] = None,
        timeout: int = 299,
        json_params: Optional[Dict] = None,
        proxy_string: Optional[Text] = None) -> Generator[Dict[Text, Any], None, None]:
    """ Execute query until we have all results """

    offset = 0
    count = 0

    proxies = {
        'http': proxy_string,
        'https': proxy_string
    }

    options = {
        "headers": headers,
        "verify": False,
        "timeout": timeout,
        "proxies": proxies,
        "params": {}
    }

    while True:  # do - while offset < count
        # ARGUS uses offset in body for POST requests and request parameters for GET requests

        if method == "POST" and json_params:
            json_params["offset"] = offset
        elif method == "GET":
            options["params"]["offset"] = offset  # type: ignore

        debug("Executing search: {}, json={}, options={}".format(url, json_params, options))
        req = requests.request(method, url, json=json_params, **options)  # type:ignore

        if not req.status_code == 200:
            errmsg = "status_code: {0.status_code}: {0.content}"
            raise worker.UnknownResult(errmsg.format(req))

        res = req.json()
        data = res["data"]
        count = res.get("count", 0)

        yield from data

        debug("count={}".format(count))

        offset += len(data)

        if offset >= count:
            break



#!/usr/bin/env python

"""

config module, supports loading config from ini, environment and arguments

The configuration presedence are (from lowest to highest):
    1. argparse default
    3. ini file
    3. environment variable
    4. command line argument

# Config

Arguments are parsed in two phases. First, it will look for the argument --config argument
which can be used to specify an alternative location for the ini file. If not --config argument
is given it will look for an ini file in the following locations:

    /etc/<CONFIG_FILE_NAME>
    ~/.config/<CONFIG_ID>/<CONFIG_FILE_NAME> (or directory specified by XDG_CONFIG_HOME)

The ini file can contain a "[DEFAULT]" section that will be used for all configurations.
In addition it can have a section that corresponds with <SECTION_NAME> that for
specific cofiguration, that will over overide config from DEFAULT

# Environment variables

The configuration step will also look for environment variables in uppercase and
with "-" replaced with "_". For the example below it will lookup the following environment
variables:

    - $NUMBER
    - $BOOL
    - $STR_ARG

Example:

>>> parser = argparse.ArgumentParser("test argparse")
>>> parser.add_argument('--number', type=int, default=1)
>>> parser.add_argument('--bool', action='store_true')
>>> parser.add_argument('--str-arg')
>>> args = config.handle_args(parser, <CONFIG_ID>, <CONFIG_FILE_NAME>, <SECTION_NAME>)

"""

import argparse
import configparser
import os
from functools import partialmethod
from typing import Any, Dict, List, Optional, Text, Tuple

# Monkeypatch ArgumentParser to not allow abbrevations as those will make it
# hard to mix and match options on commandline, env and ini files
argparse.ArgumentParser.__init__ = partialmethod(argparse.ArgumentParser.__init__, allow_abbrev=False)  # type: ignore


class SectionNotFound(Exception):
    """Config file not found"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


class NotSupported(Exception):
    """Option not supported"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


def get_xdg_dir(xdg_id: Text, env_name: Text, default: Text, create: bool = False) -> Text:
    """
    Get xdg dir.

    https://specifications.freedesktop.org/basedir-spec/basedir-spec-0.6.html

    Honors $XDG_*_HOME, but fallbacks to defaults

Args:
    xdg_id [str]: directory under directory that will be used
    env_name [str]: XDG environment variable, e.g. XDG_CACHE_HOME
    env_name [str]: default directory in home directory, e.g. .cache
    create [bool]: create directory if not exists

Return path to cache_directory
    """

    home = os.environ["HOME"]

    xdg_home = os.environ.get(env_name, os.path.join(home, default))
    xdg_dir = os.path.join(xdg_home, xdg_id)

    if create and not os.path.isdir(xdg_dir):
        os.makedirs(xdg_dir)

    return xdg_dir


def get_config_dir(config_id: Text, create: bool = False) -> Text:
    """
    Get config dir.

    Honors $XDG_CONFIG_HOME, but fallbacks to ".config"

    See get_xdg_dir for details
    """

    return get_xdg_dir(config_id, "XDG_CONFIG_HOME", ".config", create)


def find_default_ini(ini_id: Text = "actworkers",
                     ini_filename: Text = "actworkers.ini") -> Optional[Text]:
    """
    Look for default ini files in /etc and ~/.config
    """

    # Order to search for confiuration files
    locations: List[Text] = [
        os.path.join(get_config_dir(ini_id), ini_filename),
        "/etc/{}".format(ini_filename)
    ]

    ini_files: List[Text] = [loc for loc in locations if os.path.isfile(loc)]

    if not ini_files:
        return None

    with open(ini_files[0]) as f:
        return f.read()


def load_ini(config_id: Text,
             config_name: Text,
             opts: Optional[List] = None) -> Tuple[Optional[configparser.ConfigParser], List]:
    """
    return config, remainder_argv
    """

    early_parser = argparse.ArgumentParser(description="configfile parser", add_help=False)
    early_parser.add_argument('--config', dest='config',
                              type=argparse.FileType('r', encoding='UTF-8'),
                              default=None,
                              help='change default configuration location')

    args, remainder_argv = early_parser.parse_known_args(opts)

    config = args.config

    if config:
        config = config.read()

    # No config file specified on command line, attempt to find
    # in default locations
    else:
        config = find_default_ini(config_id, config_name)

    if config:
        cp = configparser.ConfigParser()
        cp.read_string(config)
        return cp, remainder_argv

    return None, remainder_argv


def get_env(key: Text) -> Dict:
    """
    Get environment variable based on key
    (uppercase and replace "-" with "_")
    """
    env_key = key.replace("-", "_").upper()

    if env_key in os.environ:
        return {key: os.environ[env_key]}
    return {}


def get_default(action: argparse.Action, section: Dict, key: Text) -> Any:
    """
    Find default value for an option. This will only be used if an
    argument is not specified at the command line. The defaults will
    be found in this order and continue until a value is found:
        1. environment variable
        2. ini file
        3. argparse default

    This will only be used if the argument is not specified at the
    command line.
    """
    default = action.default
    env = get_env(key)

    # environment has higher presedence than config section
    if key in env:
        default = env[key]
    elif key in section:
        default = section[key]

    # if not env or section, keep default from argparse

    # parse true/yes as True and false/no as False for
    # action="store_true" and action="store_false"
    if action.const in (True, False) and isinstance(default, str):
        if default.lower() in ("true", "yes"):
            default = True
        elif default.lower() in ("false", "no"):
            default = False

    if action.nargs in (argparse.ZERO_OR_MORE, argparse.ONE_OR_MORE):
        default = default.split()

    # If argument type is set and default is not None, enforce type
    # Eg, for this argument specification
    # parser.add_argument('--int-arg', type=int)
    # --int-arg 2
    # will give you int(2)
    # If --int-arg is omitted, it will use None
    if action.type is not None and default is not None:
        default = action.type(default)

    return default


def handle_args(parser: argparse.ArgumentParser,
                config_id: Text,
                config_name: Text,
                section_name: Text,
                opts: Optional[List] = None) -> argparse.Namespace:
    """
    parses and sets up the command line argument system above
    with config file parsing.
    """

    cp, remainder_argv = load_ini(config_id, config_name, opts=opts)

    if cp:
        # Add (empty) section. In this way we can still access
        # the DEFAULT section
        if not cp.has_section(section_name):
            cp.add_section(section_name)
        config = dict(cp[section_name])
    else:
        config = {}

    # Loop over parser groups / actions
    # Unfortunately we can only do this in protected members..
    # pylint: disable=protected-access
    for g in parser._action_groups:
        for action in g._actions:
            if action.required:
                raise NotSupported('"required" argument is not supported (found in option {}). '.format("".join(action.option_strings)) +
                                   "Set to false and test after it has been parsed by handle_args()")
            for option_string in action.option_strings:
                if option_string.startswith('--'):
                    key=option_string[2:]
                    action.default=get_default(action, config, key)

    return parser.parse_args(remainder_argv)



""" test config """
import argparse
import os

from act.workers.libs import config

INI_TEST_FILE = os.path.join(os.path.dirname(__file__), "data/config_testdata.ini")


def __argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser("test argparse", allow_abbrev=False)
    parser.add_argument('--number', type=int, default=1)
    parser.add_argument('--bool', action='store_true')
    parser.add_argument('--str-arg')

    return parser


def test_argparse_only():
    """ all arguments from command line, using default for number and bool """

    parser = __argparser()

    commandline = "--str-arg test".split()

    args = config.handle_args(parser, "actworkers", "actworkers.ini", "test", opts=commandline)

    assert args.number == 1
    assert args.str_arg == "test"
    assert not args.bool


def test_argparse_ini():
    """ all arguments from ini file """
    parser = __argparser()

    commandline = "--config {}".format(INI_TEST_FILE).split()

    args = config.handle_args(parser, "actworkers", "actworkers.ini", "test", opts=commandline)

    assert args.number == 3
    assert args.str_arg == "from ini"
    assert args.bool is True


def test_argparse_env():
    """ all arguments from env """
    parser = __argparser()

    env = {
        "STR_ARG": "from env",
        "NUMBER": 4,
        "BOOL": "yes"  # accepts both yes and true
    }

    for key, value in env.items():
        os.environ[key] = str(value)

    args = config.handle_args(parser, "actworkers", "actworkers.ini", "test", opts=[])

    assert args.number == 4
    assert args.str_arg == "from env"
    assert args.bool is True

    # Remove from environment variables
    for key in env:
        del os.environ[key]


def test_argparse_env_ini():
    """
    --number from enviorment
    --bool from ini
    --str-arg from cmdline

    """
    parser = __argparser()

    env = {
        "NUMBER": 4,
    }

    for key, value in env.items():
        os.environ[key] = str(value)

    commandline = "--config {} --str-arg cmdline".format(INI_TEST_FILE).split()

    args = config.handle_args(parser, "actworkers", "actworkers.ini", "test", opts=commandline)

    assert args.number == 4
    assert args.str_arg == "cmdline"
    assert args.bool is True

    # Remove from environment variables
    for key in env:
        del os.environ[key]



""" Argus module to create facts from argus events """

import re
from collections import defaultdict
from ipaddress import AddressValueError, IPv4Address
from logging import debug, error, warning
from typing import Dict, Text, cast, List

import act.api
import act.api.helpers

HASH_MD5_RE = re.compile(r'^[0-9a-fA-F]{32}$')
HASH_SHA1_RE = re.compile(r'^[0-9a-fA-F]{40}$')
HASH_SHA256_RE = re.compile(r'^[0-9a-fA-F]{64}$')
HASH_SHA5112_RE = re.compile(r'^[0-9a-fA-F]{128}$')

def is_public_ip(ip_str: Text) -> bool:
    """ Return True if IP (str) is a valid, public IP """
    try:
        return IPv4Address(ip_str).is_global
    except AddressValueError:
        return False


def handle_fact(fact: act.api.fact.Fact, output_format: Text) -> None:
    """ wrap act.helpers.handle_fact and log all errors (and continue) """
    try:
        act.api.helpers.handle_fact(fact, output_format)
    except act.api.base.ResponseError:
        error("Error adding fact (ResponseError): {}".format(fact, exc_info=True))
    except act.api.schema.MissingField:
        error("Error adding fact (missing field): {}".format(fact, exc_info=True))


def handle_uri(actapi: act.api.Act, uri: Text, output_format: Text) -> None:
    """ wrap act.helpers.handle_uri and log all errors (and continue) """
    try:
        act.api.helpers.handle_uri(actapi, uri, output_format)
    except act.api.base.ResponseError:
        error("Error adding uri (ResponseError): {}".format(uri, exc_info=True))
    except act.api.schema.MissingField:
        error("Error adding uri (missing field): {}".format(uri, exc_info=True))


def handle_argus_event_hash(
        actapi: act.api.Act,
        properties: defaultdict,
        event_id: Text,
        content_props: List[Text],
        hash_props: List[Text],
        output_format: Text) -> None:
    """ Handle file/process hash properties in events """

    # For all hash values that represents a content object (typically sha256)
    for prop in content_props:
        if properties[prop]:
            # Properties can be aggregated (separated by newlines)
            for value in properties[prop].split("\n"):
                if not HASH_SHA256_RE.search(value):
                    warning('Illegal sha256: "{}" in property "{}"'.format(value, prop))
                    continue

                handle_fact(
                    actapi.fact("observedIn", "event")
                    .source("content", value)
                    .destination("event", event_id),
                    output_format=output_format
                )

    # For all hash values that needs a placeholder (unknown content/sha256)
    for prop in hash_props:
        if properties[prop]:
            for value in properties[prop].split("\n"):
                if not (HASH_MD5_RE.search(value) or HASH_SHA1_RE.search(value) or HASH_SHA5112_RE.search(value)):
                    warning("Unknown hash: {} in property {}".format(value, prop))
                    continue

                if value:
                    chain = act.api.fact.fact_chain(
                        actapi.fact("represents")
                        .source("hash", value)
                        .destination("content", "*"),
                        actapi.fact("observedIn", "event")
                        .source("content", "*")
                        .destination("event", event_id),
                    )

                    for fact in chain:
                        handle_fact(fact, output_format=output_format)


def get_scheme(event: defaultdict) -> Text:
    "Get scheme based on protocol, default to 'network'"

    protocols = ("tcp", "udp", "icmp")

    # If we have a known protocol, use this as scheme when constructing URI
    if event["protocol"] in protocols:
        return cast(Text, event["protocol"])
    return "network"


def handle_argus_event_ip(
        actapi: act.api.Act,
        event: defaultdict,
        event_id: Text,
        output_format: Text) -> None:
    "Create facts from source/destination IP address"

    scheme = get_scheme(event)

    for direction in ["source", "destination"]:
        # ensure we have IP information for this direction
        if not event[direction] and event[direction]["networkAddress"]:
            continue

        address = event[direction]["networkAddress"]["address"]

        if not is_public_ip(address):
            continue

        # Check if the flag <direction>_IS_CUSTOMERNET is set
        if "{}_IS_CUSTOMERNET".format(direction.upper()) in event["flags"]:
            debug("Address is customernet: {}, event={}".format(address, event))
            continue

        uri = "{}://{}".format(scheme, address)

        # Facts: uri components
        handle_uri(actapi, uri, output_format=output_format)

        # Fact: uri -> event
        handle_fact(
            actapi.fact("observedIn", "event").source("uri", uri).destination("event", event_id),
            output_format=output_format)


def handle_argus_event_fqdn(
        actapi: act.api.Act,
        event: defaultdict,
        event_id: Text,
        output_format: Text) -> None:
    "Create fact from fqdn"

    scheme = get_scheme(event)

    # Fact: fqdn -> event
    if event["domain"] and event["domain"]["fqdn"]:
        uri = "{}://{}".format(scheme, event["domain"]["fqdn"])

        # Facts, uri components
        handle_uri(actapi, uri, output_format=output_format)

        # Fact: uri -> event
        handle_fact(
            actapi.fact("observedIn", "event").source("uri", uri).destination("event", event_id),
            output_format=output_format)


def handle_argus_event(
        actapi: act.api.Act,
        event: Dict,
        content_props: List[Text],
        hash_props: List[Text],
        output_format: Text) -> None:
    "handle all facts from an argus event"

    # Use defaultdicts that defaults to None
    properties = defaultdict(lambda: None, event["properties"])
    event = defaultdict(lambda: None, event)

    signature = event["attackInfo"]["signature"]

    # Use argus ID (AGGRid) as event_id
    event_id = "ARGUS-{}".format(event["id"])

    # Fact: event -> incident
    if event["associatedCase"]:
        case_id = "ARGUS-{}".format(event["associatedCase"]["id"])
        handle_fact(
            actapi.fact("attributedTo", "incident").source("event", event_id).destination("incident", case_id),
            output_format=output_format)
        handle_fact(
            actapi.fact("name", event["associatedCase"]["description"]).source("incident", case_id),
            output_format=output_format)

    # Facts: hash/content -> event
    handle_argus_event_hash(actapi, properties, event_id, content_props, hash_props, output_format)

    # Facts: sourceIP -> event, destinationIP -> event
    handle_argus_event_ip(actapi, event, event_id, output_format)

    # Fact: signature -> event
    handle_fact(
        actapi.fact("detects", "event").source("signature", signature).destination("event", event_id),
        output_format=output_format)

    # Fact: uri -> event
    if event["uri"]:
        handle_uri(actapi, event["uri"], output_format=output_format)
        handle_fact(
            actapi.fact("observedIn", "event").source("uri", event["uri"]).destination("event", event_id),
            output_format=output_format)
    else:
        # Only construct URI from fqdn if we do not have an uri on the event. The URI
        # Is normally more correct (e.g. scheme and path will be more correctly specified)
        # Facts: uri fqdn -> uri ->  event
        handle_argus_event_fqdn(actapi, event, event_id, output_format)



import pytest
from act.workers.libs import misp
import json
import uuid


def test_event():
    try:
        from StringIO import StringIO
    except ImportError:
        from io import StringIO

    a = misp.Event()
    output = StringIO()
    with pytest.raises(AttributeError):
        a.uuid = "GURBA"

    a.write_to(output)
    event = {"Event": json.loads(output.getvalue())}

    b = misp.Event(loads=json.dumps(event))

    assert a.uuid == b.uuid


def test_load_event():
    with open("test/misp_event.json") as event_f:
        a = misp.Event(loads=event_f.read())
    assert a.info == 'M2M -  Dridex 2017-11-14 : botnet 7200 : "Invoice No.\n 123456" - "Invoice-123456-06.doc"'
    assert a.uuid == uuid.UUID("5a0f0fb1-0b54-4ace-bb7b-429f950d210f")
    assert a.date == "2017-11-17"
    assert a.published is False
    assert a.threat_level_id == misp.ThreatLevelID.LOW
    assert a.analysis == misp.Analysis.ONGOING
    assert "tool" in a.misp_galaxy
    assert a.misp_galaxy.get("tool", "") == "Dridex"



#!/usr/bin/env python3

"""

Worker for Mitre ATT&CK, using the STIX implementation available here:

    https://github.com/mitre/cti

    ATT&CK Property     STIX Object type        ACT object
    =========================================================
    Technique           attack-pattern          technique
    Group               intrusion-set           threatActor
    Software	        malware or tool         tool
    Mitigation	        course-of-action        n/a

"""

import argparse
import os
import sys
import traceback
from logging import error, info, warning
from typing import Any, Dict, List, Text

import stix2
from stix2 import Filter, MemoryStore, parse

import act.api
from act.api.helpers import handle_fact
from act.workers.libs import worker

MITRE_URLS = {
    "enterprise": "https://raw.githubusercontent.com/mitre/cti/master/enterprise-attack/enterprise-attack.json",
    "pre": "https://raw.githubusercontent.com/mitre/cti/master/pre-attack/pre-attack.json",
    "mobile": "https://raw.githubusercontent.com/mitre/cti/master/mobile-attack/mobile-attack.json"
}

DEFAULT_NOTIFY_CACHE = os.path.join(os.environ["HOME"], "act-mitre-attack-notify.cache")


class NotificationError(Exception):
    """NotificationError"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


def parseargs() -> argparse.ArgumentParser:
    """ Parse arguments """
    parser = worker.parseargs('Mitre ATT&CK worker')
    parser.add_argument('--smtphost', dest='smtphost', help="SMTP host used to send revoked/deprecated objects")
    parser.add_argument('--sender', dest='sender', help="Sender address used to send revoked/deprecated objects")
    parser.add_argument('--recipient', dest='recipient', help="Recipient address used to send revoked/deprecated objects")
    parser.add_argument(
        '--type',
        choices=list(MITRE_URLS.keys()),
        help='Specify a single type to download (enterprise, mobile or pre). Default is to fetch all')
    parser.add_argument('--notifycache', dest='notifycache', help="Cache for revoked/deprecated objects", default=DEFAULT_NOTIFY_CACHE)

    return parser


def get_attack(url: str, proxy_string: str, timeout: int) -> MemoryStore:
    """Fetch Mitre ATT&CK JSON data in Stix2 format and return a Stix2 memory store"""
    attack = worker.fetch_json(url, proxy_string, timeout)

    # Create memory store
    mem = MemoryStore()

    # Add all objects to the memory store
    for obj in parse(attack, allow_custom=True).objects:
        mem.add(obj)

    return mem


def add_techniques(client, attack: MemoryStore, output_format: Text = "json") -> List[stix2.AttackPattern]:
    """
        extract objects/facts related to ATT&CK techniques

    Args:
        attack (stix2):       Stix attack instance

    """

    notify = []

    # ATT&CK concept    STIX Object type        ACT object
    # =========================================================
    # Technique         attack-pattern          technique
    # Filter out ATT&CK techniques (attack-pattern) from bundle

    for technique in attack.query([Filter("type", "=", "attack-pattern")]):
        if getattr(technique, "revoked", None):
            # Object is revoked, add to notification list but do not add to facts that should be added to the platform
            notify.append(technique)
            continue

        if getattr(technique, "x_mitre_deprecated", None):
            # Object is revoked, add to notification list AND continue to add to facts that should be added to the platform
            notify.append(technique)

        # Mitre ATT&CK Tactics are implemented in STIX as kill chain phases with kill_chain_name "mitre-attack"
        for tactic in technique.kill_chain_phases:
            if tactic.kill_chain_name != "mitre-attack":
                continue

            handle_fact(
                client.fact("accomplishes")
                .source("technique", technique.name)
                .destination("tactic", tactic.phase_name),
                output_format=output_format
            )

    return notify


def add_groups(client, attack: MemoryStore, output_format: Text = "json") -> List[stix2.AttackPattern]:
    """
        extract objects/facts related to ATT&CK Groups

    Args:
        attack (stix2):       Stix attack instance

    """

    notify = []

    # ATT&CK concept    STIX Object type        ACT object
    # =========================================================
    # Group	        intrusion-set           threatActor
    #
    # Filter out ATT&CK groups (intrusion-set) from bundle

    for group in attack.query([Filter("type", "=", "intrusion-set")]):
        if getattr(group, "revoked", None):
            # Object is revoked, add to notification list but do not add to facts that should be added to the platform
            notify.append(group)
            continue

        if getattr(group, "x_mitre_deprecated", None):
            # Object is revoked, add to notification list AND continue to add to facts that should be added to the platform
            notify.append(group)

        for alias in getattr(group, "aliases", []):
            if group.name != alias:
                handle_fact(
                    client.fact("alias")
                    .bidirectional("threatActor", group.name, "threatActor", alias),
                    output_format=output_format
                )

        #   ATT&CK concept   STIX Properties
        #   ==========================================================================
        #   Software         relationship where relationship_type == "uses",
        #                    points to a target object with type== "malware" or "tool"

        for tool in attack.related_to(group, relationship_type="uses"):
            if tool.type not in ("malware", "tool"):
                continue

            chain = act.api.fact.fact_chain(
                client.fact("classifiedAs")
                .source("content", "*")
                .destination("tool", tool.name.lower()),
                client.fact("observedIn", "event")
                .source("content", "*")
                .destination("event", "*"),
                client.fact("attributedTo", "incident")
                .source("event", "*")
                .destination("incident", "*"),
                client.fact("attributedTo", "threatActor")
                .source("incident", "*")
                .destination("threatActor", group.name)
            )

            for fact in chain:
                handle_fact(fact, output_format=output_format)

        #   ATT&CK concept   STIX Properties
        #   ==========================================================================
        #   Technqiues       relationship where relationship_type == "uses", points to
        #                    a target object with type == "attack-pattern"

        for technique in attack.related_to(group, relationship_type="uses"):
            if technique.type != "attack-pattern":
                continue

            chain = act.api.fact.fact_chain(
                client.fact("classifiedAs", "technique")
                .source("event", "*")
                .destination("technique", technique.name),
                client.fact("attributedTo", "incident")
                .source("event", "*")
                .destination("incident", "*"),
                client.fact("attributedTo")
                .source("incident", "*")
                .destination("threatActor", group.name)
            )

            for fact in chain:
                handle_fact(fact, output_format=output_format)

    return notify


def add_software(client, attack: MemoryStore, output_format: Text = "json") -> List[stix2.AttackPattern]:
    """
        extract objects/facts related to ATT&CK Software
        Insert to ACT if client.baseurl is set, if not, print to stdout

    Args:
        attack (stix2):       Stix attack instance

    """

    notify = []

    for software in attack.query([Filter("type", "in", ["tool", "malware"])]):
        tool_name = software.name.lower()

        # Tool category
        handle_fact(
            client.fact("category", software.type).source("tool", tool_name),
            output_format=output_format
        )

        if getattr(software, "revoked", None):
            # Object is revoked, add to notification list but do not add to facts that should be added to the platform
            notify.append(software)
            continue

        if getattr(software, "x_mitre_deprecated", None):
            # Object is revoked, add to notification list AND continue to add to facts that should be added to the platform
            notify.append(software)

        for alias in getattr(software, "x_mitre_aliases", []):
            if tool_name != alias.lower():
                # Tool category (alias)
                handle_fact(
                    client.fact("category", software.type).source("tool", alias.lower()),
                    output_format=output_format
                )
                handle_fact(
                    client.fact("alias")
                    .bidirectional("tool", tool_name, "tool", alias.lower()),
                    output_format=output_format
                )

        #   ATT&CK concept   STIX Properties
        #   ==========================================================================
        #   Technqiues       relationship where relationship_type == "uses", points to
        #                    a target object with type == "attack-pattern"

        for technique in attack.related_to(software, relationship_type="uses"):
            if technique.type != "attack-pattern":
                continue

            handle_fact(
                client.fact("implements")
                .source("tool", software.name.lower())
                .destination("technique", technique.name),
                output_format=output_format
            )

    return notify


def notify_cache(filename: str) -> Dict:
    """
    Read notify cache from filename
    Args:
        filename(str):      Cache filename

    """

    cache = {}

    try:
        with open(filename) as f:
            for line in f:
                if line:
                    cache[line.strip()] = True
    except FileNotFoundError:
        warning("Cache file {} not found, will be created if necessary".format(filename))

    return cache


def add_to_cache(filename: str, entry: str) -> None:
    """
    Add entry to cache

    Args:
        filename(str):      Cache filename
        entry(str):         Cache entry
    """

    with open(filename, "a") as f:
        f.write(entry.strip())
        f.write("\n")


def send_notification(
        notify: List[stix2.AttackPattern],
        smtphost: str,
        sender: str,
        recipient: str,
        url: str) -> List[str]:
    """
    Process revoked objects

    Args:
        notify(attack[]):   Array of revoked/deprecated Stix objects
        notifycache(str):   Filename of notify cache
        smtphost(str):      SMTP host used to notify of revoked/deprecated objects
        sender(str):        sender address used to notify of revoked/deprecated objects
        recipient(str):     recipient address used to notify of revoked/deprecated objects

    smtphost, sender AND recipient must be set to notify of revoked/deprecated objects

    Return list of IDs that was successfully notified

    """

    notified = []

    if not (smtphost and recipient and sender):
        error("--smtphost, --recipient and --sender must be set to send revoked/deprecated objects on email")
        return []

    body = url + "\n\n"
    warning("[{}]".format(url))

    for obj in notify:
        if getattr(obj, "revoked", None):
            text = "revoked: {}:{}".format(obj.type, obj.name)

        elif getattr(obj, "x_mitre_deprecated", None):
            text = "deprecated: {}:{}".format(obj.type, obj.name)

        else:
            raise NotificationError("object tis not deprecated or revoked: {}:{}".format(obj.type, obj.name))

        notified.append(obj.id)

        body += text + "\n"
        warning(text)

    worker.sendmail(smtphost, sender, recipient, "Revoked/deprecated objects from MITRE/ATT&CK", body)
    info("Email sent to {}".format(recipient))

    return notified


def main() -> None:
    """ Main function """

    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = worker.handle_args(parseargs())

    actapi = worker.init_act(args)

    if args.type:
        types = [args.type]
    else:
        types = list(MITRE_URLS.keys())

    for mitre_type in types:
        url = MITRE_URLS.get(mitre_type.lower())

        if not url:
            error("Unknown mitre type: {}. Valid types: {}".format(mitre_type, ",".join(MITRE_URLS.keys())))
            sys.exit(2)

        cache = notify_cache(args.notifycache)

        # Get attack dataset as Stix Memory Store
        attack = get_attack(url, args.proxy_string, args.http_timeout)

        techniques_notify = add_techniques(actapi, attack, args.output_format)
        groups_notify = add_groups(actapi, attack, args.output_format)
        software_notify = add_software(actapi, attack, args.output_format)

        # filter revoked objects from those allready notified
        notify = [
            notify
            for notify in techniques_notify + groups_notify + software_notify
            if notify.id not in cache
        ]

        if notify:
            notified = send_notification(notify, args.smtphost, args.sender, args.recipient, url)

            for object_id in notified:
                # Add object to cache, so we will not be notified on the same object on the next run
                add_to_cache(args.notifycache, object_id)


def main_log_error() -> None:
    "Call main() and log all excetions  as errors"
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == '__main__':
    main_log_error()



"""Common worker library"""

import argparse
import inspect
import os
import smtplib
import socket
from email.mime.text import MIMEText
from logging import error
from typing import Any, Optional, Text

import requests
import urllib3

import act.api
from act.workers.libs import config

CONFIG_ID = "actworkers"
CONFIG_NAME = "actworkers.ini"


class UnknownResult(Exception):
    """UnknownResult is used in API request (not 200 result)"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


def parseargs(description: str) -> argparse.ArgumentParser:
    """ Parse arguments """
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="{} ({})".format(description, worker_name()), epilog="""

  --config INI_FILE     Override default locations of ini file

    Arguments can be specified in ini-files, environment variables and
    as command line arguments, and will be parsed in that order.

    By default, workers will look for an ini file in /etc/{1}
    and ~/.config/{0}/{1} (or in $XDG_CONFIG_DIR if
    specified).

    Each worker will read the confiuration from the "DEFAULT" section in the
    ini file, and in it's own section (in that order).

    It is also possible to use environment variables for configuration.
    Workers will look for environment variables for all arguments with
    the argument name in uppercase and "-" replaced with "_".

    E.g. set the CERT_FILE environment variable to configure the
    --cert-file option.

""".format(CONFIG_ID, CONFIG_NAME), formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('--http-timeout', dest='http_timeout', type=int,
                        default=120, help="Timeout")
    parser.add_argument('--proxy-string', dest='proxy_string', help="Proxy to use for external queries")
    parser.add_argument('--cert-file', dest='cert_file', help="Cerfiticate to add if you are behind a SSL/TLS interception proxy.")
    parser.add_argument('--user-id', dest='user_id',
                        help="User ID")
    parser.add_argument('--act-baseurl', dest='act_baseurl',
                        help='ACT API URI')
    parser.add_argument("--logfile", dest="logfile",
                        help="Log to file (default = stdout)")
    parser.add_argument("--loglevel", dest="loglevel", default="info",
                        help="Loglevel (default = info)")
    parser.add_argument("--output-format", dest="output_format", choices=["str", "json"], default="json",
                        help="Output format for fact (default=json)")
    parser.add_argument('--http-user', dest='http_user', help="ACT HTTP Basic Auth user")
    parser.add_argument('--http-password', dest='http_password', help="ACT HTTP Basic Auth password")
    return parser


def __mod_name(stack: inspect.FrameInfo) -> Text:
    """ Return name of module from a stack ("_" is replaced by "-") """
    mod = inspect.getmodule(stack[0])
    return os.path.basename(mod.__file__).replace(".py", "").replace("_", "-")


def worker_name() -> Text:
    """ Return first external module that called this function, directly, or indirectly """

    modules = [__mod_name(stack) for stack in inspect.stack() if __mod_name(stack)]
    return [name for name in modules if name != modules[0]][0]


def handle_args(parser: argparse.ArgumentParser) -> argparse.Namespace:
    """ Wrapper for config.handle_args where we set config_id and config_name """
    return config.handle_args(parser, CONFIG_ID, CONFIG_NAME, worker_name())


def init_act(args: argparse.Namespace) -> act.api.Act:
    """ Initialize act api from arguments """
    requests_kwargs = {}
    if args.http_user:
        requests_kwargs["auth"] = (args.http_user, args.http_password)

    if args.cert_file:
        requests_kwargs["verify"] = args.cert_file

    return act.api.Act(
        args.act_baseurl,
        args.user_id,
        args.loglevel,
        args.logfile,
        worker_name(),
        requests_common_kwargs=requests_kwargs)


def get_cache_dir(cache_id: str, create: bool = False) -> str:
    """
    Get cache dir.

    Honors $XDG_CACHE_HOME, but fallbacks to $HOME/.cache

Args:
    cache_id [str]: directory under CACHE that will be used
    create [bool]: create directory if not exists

Return path to cache_directory
    """

    home = os.environ["HOME"]

    cache_home = os.environ.get("XDG_CACHE_HOME", os.path.join(home, ".cache"))
    cache_dir = os.path.join(cache_home, cache_id)

    if create and not os.path.isdir(cache_dir):
        os.makedirs(cache_dir)

    return cache_dir


def fetch_json(url: str, proxy_string: Optional[str], timeout: int = 60, verify_https: bool = False) -> Any:
    """Fetch remote URL as JSON
    url (string):                    URL to fetch
    proxy_string (string, optional): Optional proxy string on format host:port
    timeout (int, optional):         Timeout value for query (default=60 seconds)
    """

    proxies = {
        'http': proxy_string,
        'https': proxy_string
    }

    options = {
        "verify": verify_https,
        "timeout": timeout,
        "proxies": proxies,
        "params": {}
    }

    if not verify_https:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    try:
        req = requests.get(url, **options)
    except (urllib3.exceptions.ReadTimeoutError,
            requests.exceptions.ReadTimeout,
            socket.timeout) as err:
        error("Timeout ({0.__class__.__name__}), query: {1}".format(err, req.url))

    if not req.status_code == 200:
        errmsg = "status_code: {0.status_code}: {0.content}"
        raise UnknownResult(errmsg.format(req))

    return req.json()


def sendmail(smtphost: str, sender: str, recipient: str, subject: str, body: str) -> None:
    """Send email"""

    msg = MIMEText(body, "plain", "utf-8")
    msg['Subject'] = subject
    msg['From'] = sender
    msg['To'] = recipient
    s = smtplib.SMTP(smtphost)
    s.sendmail(sender, [recipient], msg.as_string())
    s.quit()



# -*- coding: utf-8 -*-
from collections import defaultdict
import act


def create_api_agent(baseurl, userid):
    """ creates an ACT API agent """
    c = act.Act(baseurl, userid, log_level="info")
    return c


def get_all_ta_from_act(baseurl, userid):
    """ gets set of all threat actor names in the ACT platform.
    Note that limit is set to 1000. """

    objects = create_api_agent(baseurl,
                               userid).object_search(object_type=["threatActor"], limit=1000)

    ta_set = set()

    for x in objects:
        ta_set.add(x.value)

    return list(ta_set)


def get_all_alias_facts_from_act(baseurl, userid):
    """ gets set of all bindings between threat actor and threat actor aliases
    from ACT. Note that limit is set to 1000. """

    facts = create_api_agent(baseurl, userid).\
        fact_search(object_type=["threatActor"],
                    fact_type=["threatActorAlias"], limit=1000)

    ta_set_facts = set()
    for fa in facts:
        ta1, ta2 = fa.source_object.value, fa.destination_object.value
        ta_set_facts.add((ta1, ta2))

    return ta_set_facts


def add_ta_to_map(ta_set):
    """ adds all threat actor names given in a set to a map.
    The map is a defaultdict with sets, each set containing a key and all
    relevant aliases for that key as values. """

    ta_map = defaultdict(set)
    for ta in ta_set:
        ta_map[ta].add(ta)
    return ta_map


def add_ta_alias_to_map(ta_aliases, ta_map):
    """ adds alias to the ta_map. Needs arguments ta_map defaultdict with sets,
    and tuples with two strings in each. Assumes that all threat actors are
    already in the ta_map. """

    for tup in ta_aliases:
        ta1, ta2 = tup
        s = ta_map[ta1]
        s.update(ta_map[ta2])
        # point key of all elements of the set to the same set.
        for x in s:
            ta_map[x] = s

    return ta_map


def decide_on_key(k_decide, v_decide, config_dict):
    """ checks with old config file to decide on key in new config file. """

    # if key is within the keys in the current configfile,
    # then return that value from v as key and the rest as aliases.
    if k_decide in config_dict[0]:
        v_decide.remove(k_decide)
        return k_decide, v_decide

    # if the key is within the aliases of another key, then return the key of
    # the set where the value was found and remove that value from the set.
    for kk, vv in config_dict.items():
        if k_decide in vv and kk in v_decide:
            v_decide.remove(kk)

            return kk, v_decide

    # if the key is not a key already, then just choose one from the set v,
    # and set the rest as aliases.
    v_decide.remove(k_decide)

    return k_decide, v_decide


def create_config(ta_map, aliasfile, newaliasfile):
    """ creates config file from ta_map, defaultdict with set."""
    def config_split(l):
        k, v = l.split(":")
        return k, [x.strip() for x in v.split(",")]

    config_dict = defaultdict(set)

    with open(aliasfile, "r") as config:
        for l in config:
            k, v = config_split(l)
            config_dict[k] = v

    with open(newaliasfile, "w") as config:

        config.truncate()

        while ta_map:
            k, v = ta_map.popitem()

            if len(v) == 1:
                config.write(k + ":\n")
            else:
                k_decide_key = k[:]
                v_decide_key = set(v)
                ta_name, ta_aliases = decide_on_key(k_decide_key, v_decide_key,
                                                    config_dict)
                config.write("{}: {}\n".format(ta_name, ",".join(ta_aliases)))
                v.remove(k)
                for x in v:
                    ta_map.pop(x)



"""Setup script for the python-act library module"""

from os import path

from setuptools import setup

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), "rb") as f:
    long_description = f.read().decode('utf-8')

setup(
    name="act-workers",
    version="0.7.3",
    author="mnemonic AS",
    zip_safe=True,
    author_email="opensource@mnemonic.no",
    description="ACT workers",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="MIT",
    keywords="ACT, mnemonic",
    entry_points={
        'console_scripts': [
            'act-argus-case = act.workers.argus_case:main_log_error',
            'act-vt = act.workers.vt:main_log_error',
            'act-attack = act.workers.attack:main_log_error',
            'act-mnemonic-pdns = act.workers.mnemonic_pdns:main_log_error',
            'act-country-regions = act.workers.country_regions:main_log_error',
            'act-misp-feeds = act.workers.misp_feeds:main_log_error',
            'act-scio = act.workers.scio:main_log_error',
            'act-uploader = act.workers.generic_uploader:main_log_error',
            'act-shadowserver-asn = act.workers.shadowserver_asn:main_log_error',
            'act-worker-config = act.workers.worker_config:main',
        ]
    },

    # Include ini-file(s) from act/workers/etc
    package_data={'act.workers': ['etc/*.ini']},
    packages=["act.workers", "act.workers.libs"],

    # https://packaging.python.org/guides/packaging-namespace-packages/#pkgutil-style-namespace-packages
    # __init__.py under all packages under in the act namespace must contain exactly string:
    # __path__ = __import__('pkgutil').extend_path(__path__, __name__)
    namespace_packages=['act'],
    url="https://github.com/mnemonic-no/act-workers",
    install_requires=['act-api>=0.6.0,<0.7.0', 'requests', 'RashlyOutlaid', 'virustotal-api', 'stix2'],
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: OSI Approved :: ISC License (ISCL)",
    ],
)



#!/usr/bin/env python3

'''Shadowserver ASN worker for the ACT platform

Copyright 2019 the ACT project <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
'''

import argparse
import json
import os
import re
import socket
import sqlite3
import sys
import time
import traceback
from ipaddress import AddressValueError, IPv4Address
from logging import debug, error, info, warning
from typing import Dict, Generator, List, Text, Tuple, Union

from RashlyOutlaid.libwhois import ASNRecord, ASNWhois, QueryError

import act.api
from act.api.helpers import handle_fact
from act.workers.libs import worker

CACHE_DIR = worker.get_cache_dir("shadowserver-asn-worker", create=True)
VERSION = "0.1"
ISO_3166_FILE = "https://raw.githubusercontent.com/lukes/" + \
    "ISO-3166-Countries-with-Regional-Codes/master/all/all.json"

# Blacklists of IPs record values
# If value matches blacklist it should not be used
BLACKLIST = {
    "ip": [  # Blacklist IP addresses. Values is IP
        lambda ip: not ip.strip(),                             # Empty values
        lambda ip: ip.strip().lstrip("0").startswith("."),     # IP addreses starting with "0."
        lambda ip: ip == "255.255.255.255",                    # broadcast
        lambda ip: IPv4Address(ip).is_multicast,
        lambda ip: IPv4Address(ip).is_private,
        lambda ip: IPv4Address(ip).is_loopback,
        lambda ip: IPv4Address(ip).is_unspecified,
    ],
    "isp": [  # Blacklist ISPs. Values is asn_record
        lambda asn_record: not asn_record.isp.strip(),         # Exclude Empty values
        lambda asn_record: asn_record.isp == asn_record.cn,    # Exclude values where ISP name == Country Name
        lambda asn_record: asn_record.isp == "AS, {}".format(asn_record.cn),  # Exclude values where ISP name == AS, <CN>
        lambda asn_record: asn_record.isp == ", {}".format(asn_record.cn)     # Exclude values where ISP name == , <CN>
    ],
    "asname": [  # Blacklist ASNAMES. Values is asn_record
        lambda asn_record: not asn_record.asname.strip(),         # Exclude Empty values
    ],
    "cn": [  # Blacklist ASNAMES. Values is asn_record
        lambda asn_record: not asn_record.cn.strip(),         # Exclude Empty values
    ]
}


def get_cn_map(filename: str) -> Dict:
    """
    Read file with county information (ISO 3166 from filename)
    return map with country code (e.g. "NO") as key, and Country
    Name (e.g. "Norway" as value)
    """
    cn_map = {}

    for c_map in json.loads(open(filename, "r", encoding="utf8").read()):
        cn_map[c_map["alpha-2"]] = c_map["name"]

    return cn_map


def parseargs() -> argparse.ArgumentParser:
    """ Parse arguments """
    parser = worker.parseargs('Shadowserver ASN enrichment')
    parser.add_argument(
        '--country-codes',
        help="Should point to file downloaded from {}".format(ISO_3166_FILE))

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--bulk', help='bulk query from file. File must contain one IP per line.')
    group.add_argument('--stdin', action='store_true', help='query ip on stdin')
    return parser


def blacklisted(value: Union[ASNRecord, str], blacklist_type: str) -> bool:
    """ Return true if value is blacklisted for the specified type """
    return any([b(value) for b in BLACKLIST[blacklist_type]])  # type: ignore


def get_db_cache(cache_dir: str) -> sqlite3.Connection:
    """
    Open cache and return sqlite3 connection
    Table is created if it does not exists
    """
    cache_file = os.path.join(cache_dir, "cache.sqlite3")
    conn = sqlite3.connect(cache_file)
    cursor = conn.cursor()
    cursor.execute("""CREATE TABLE IF NOT EXISTS asn (
        ip string unique,
        asn int,
        prefix string,
        asname string,
        cn string,
        isp string,
        peers string,
        added int)
    """)
    cursor.execute("CREATE INDEX IF NOT EXISTS asn_ip on ASN(ip)")

    return conn


def query_cache(cache: sqlite3.Connection, ip_list: List[str]) -> Generator[Tuple[str, ASNRecord], None, None]:
    """ Query cache for all IPs in list """
    cursor = cache.cursor()

    in_list = ",".join(['"{}"'.format(ip) for ip in ip_list])

    for res in cursor.execute("SELECT * FROM asn WHERE ip in ({})".format(in_list)).fetchall():
        asn_tuple = list(res[1:7])
        asn_tuple[5] = str(asn_tuple[5]).split(",")  # Split peers into list
        yield (res[0], ASNRecord(*asn_tuple))


def add_to_cache(cache: sqlite3.Connection, ip: str, asn_record: ASNRecord) -> None:
    """ ADD IP/ASNRecord to cache """
    cursor = cache.cursor()

    # flatten peer list to comma separated list
    asn_flattened = list(asn_record)
    asn_flattened[5] = ",".join(asn_record.peers)

    cursor.execute("INSERT INTO asn VALUES (?,?,?,?,?,?,?,?)",
                   ([ip] + list(asn_flattened) + [int(time.time())]))
    cache.commit()


def asn_query(ip_list: List[str], cache: sqlite3.Connection) -> Generator[Tuple[str, ASNRecord], None, None]:
    """
    Query shadowserver ASN usingi IP
    Return cached result if the IP is in the cace

    Returns tupe og IP and ASNRecord
    """

    query_ip = set(ip_list)

    for (ip, asn_record) in query_cache(cache, ip_list):
        info("Result from cache: {}".format(asn_record))
        yield (ip, asn_record)
        # Do not query IP, since we found it in cache
        query_ip.remove(ip)

    asnwhois = ASNWhois()
    asnwhois.query = list(query_ip)
    asnwhois.peers = True

    for ip in query_ip:
        try:
            asn_record = asnwhois.result[ip]
        except QueryError:
            error("Query error: {}".format(traceback.format_exc()))
            continue
        except KeyError:
            error("Key error: {}: {}".format(ip, traceback.format_exc()))
            continue
        except socket.timeout:
            error("Socket timeout query shadowserver asn: {} ({}".format(ip, query_ip))
            break  # This will also lead to timeout on all other IPs in this bulk query

        if not asn_record.asn:
            warning("No ASN found for ip {}".format(ip))
            continue

        info("Result from query: {}".format(asn_record))

        add_to_cache(cache, ip, asn_record)

        yield (ip, asn_record)


def handle_ip(
        actapi: act.api.Act,
        cn_map: Dict[str, str],
        ip_list: List[str],
        cache: sqlite3.Connection,
        output_format: Text = "json") -> None:
    """
    Read ip from stdin and query shadowserver - asn.
    if actapi is set, result is added to the ACT platform,
    if not the result is output to stdout.
    """

    ip_query = []

    for ip in ip_list:
        try:
            ip_str = str(IPv4Address(ip))
        except AddressValueError:
            error("Illegal IP address: {}".format(ip))
            continue

        # Exclude blacklisted IPs
        if blacklisted(ip_str, "ip"):
            warning("IP address is blacklisted: {}".format(ip))
        else:
            ip_query.append(ip_str)

    for (ip, res) in asn_query(ip_query, cache):
        # Remove everything after first occurence of "," in isp name
        handle_fact(
            actapi.fact("memberOf", "ipv4Network")
            .source("ipv4", ip)
            .destination("ipv4Network", res.prefix),
            output_format=output_format
        )
        handle_fact(
            actapi.fact("memberOf", "asn")
            .source("ipv4Network", res.prefix)
            .destination("asn", res.asn),
            output_format=output_format
        )

        if blacklisted(res, "asname"):
            debug('asname "{}" for ip {} is blacklisted, skipping'.format(res.asn, ip))
        else:
            handle_fact(actapi.fact("name", res.asname).source("asn", res.asn), output_format=output_format)

        if blacklisted(res, "isp"):
            debug('isp "{}" for ip {} is blacklisted, skipping'.format(res.isp, ip))
        else:
            organization = re.sub(r",.*", "", res.isp).lower()
            handle_fact(
                actapi.fact("owns", "asn")
                .source("organization", organization)
                .destination("asn", res.asn),
                output_format=output_format
            )

            if blacklisted(res, "cn"):
                debug('cn "{}" for ip {} is blacklisted, skipping'.format(res.cn, ip))
            elif res.cn not in cn_map:
                warning('Unknown cn "{}" for ip {}'.format(res.cn, ip))
            else:
                handle_fact(
                    actapi.fact("locatedIn")
                    .source("organization", organization)
                    .destination("country", cn_map[res.cn]),
                    output_format=output_format
                )


def main() -> None:
    """main function"""

    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = worker.handle_args(parseargs())

    actapi = worker.init_act(args)

    if not args.country_codes:
        sys.stderr.write("You must specify --country-codes on command line or in config file\n")
        sys.exit(1)

    if not os.path.isfile(args.country_codes):
        sys.stderr.write("Country/region file not found at specified location: {}\n".format(args.country_codes))
        sys.exit(2)

    # Get map of CC -> Country Name
    cn_map = get_cn_map(args.country_codes)

    db_cache = get_db_cache(CACHE_DIR)

    # Read IPs from stdin
    if args.stdin:
        in_data = [ip for ip in sys.stdin.read().split("\n")]
        handle_ip(actapi, cn_map, in_data, db_cache, args.output_format)

    # Bulk lookup
    elif args.bulk:
        all_ips = [ip for ip in open(args.bulk, "r")]
        batch_size = 50
        i = 0
        while i < len(all_ips):
            handle_ip(actapi, cn_map, (all_ips[i:i + batch_size]), db_cache, args.output_format)
            i += batch_size
            time.sleep(1)

    db_cache.close()


def main_log_error() -> None:
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == '__main__':
    main_log_error()
